from genericpath import samefile
import json
import logging
from reliquery.metadata import Metadata, MetadataDB, RelicTag
from typing import List, Dict
from sys import getsizeof
import os
from io import BytesIO

import numpy as np
import datetime as dt


from .storage import (
    get_all_available_storages,
    StorageItemDoesNotExist,
    get_storage_by_name,
    Storage,
)


StoragePath = List[str]

dt_format = "%m/%d/%Y %H:%M:%S"


class InvalidRelicId(Exception):
    pass


class EmptyTagDict(Exception):
    pass


class Relic:
    name: str
    relic_type: str
    storage: Storage

    def __init__(
        self,
        name: str,
        relic_type: str,
        storage: Storage = None,
        storage_name: str = "default",
        check_exists: bool = True,
    ):
        self.name = name
        self.relic_type = relic_type
        self.storage_name = storage_name

        if storage is None:
            self.storage = get_storage_by_name(self.storage_name)

        else:
            self.storage = storage

        if check_exists:
            self._ensure_exists()

        self.metadata_db = MetadataDB()

    @classmethod
    def assert_valid_id(cls, id: str):
        if id == "":
            raise InvalidRelicId()

    def _ensure_exists(self):
        try:
            self.storage.get_text([self.relic_type, self.name, "exists"])
        except StorageItemDoesNotExist as e:
            logging.info("Creating a Relic")
            self.storage.put_text([self.relic_type, self.name, "exists"], "exists")

    # TODO: needs test coverage
    @classmethod
    def relic_exists(
        cls,
        name: str,
        relic_type: str,
        storage: Storage = None,
        storage_name: str = "default",
    ) -> bool:
        if storage is None:
            storage = get_storage_by_name(storage_name)
        try:
            storage.get_text([relic_type, name, "exists"])
        except StorageItemDoesNotExist:
            return False

        return True

    def add_array(self, name: str, array: np.ndarray):

        self.assert_valid_id(name)

        size = getsizeof(array)
        shape = str(np.array(array).shape)
        metadata = Metadata(
            name=name,
            data_type="arrays",
            relic_type=self.relic_type,
            storage_name=self.storage.name,
            size=size,
            shape=shape,
        )

        buffer = BytesIO()
        np.save(buffer, array, allow_pickle=False)
        buffer.seek(0)

        self.storage.put_binary_obj(
            [self.relic_type, self.name, "arrays", name], buffer
        )
        self._add_metadata(metadata)

    def get_array(self, name: str) -> np.ndarray:
        self.assert_valid_id(name)

        with self.storage.get_binary_obj(
            [self.relic_type, self.name, "arrays", name]
        ) as f:
            return np.load(f, allow_pickle=False)

    def list_arrays(self) -> List[np.ndarray]:
        return self.storage.list_keys([self.relic_type, self.name, "arrays"])

    def add_html(self, name: str, html_path: str) -> None:

        self.assert_valid_id(name)

        metadata = metadata = Metadata(
            name=name,
            data_type="html",
            relic_type=self.relic_type,
            storage_name=self.storage.name,
        )

        self.storage.put_file([self.relic_type, self.name, "html", name], html_path)
        self._add_metadata(metadata)

    def list_html(self) -> List[str]:
        return self.storage.list_keys([self.relic_type, self.name, "html"])

    def get_html(self, name: str) -> str:
        self.assert_valid_id(name)

        with self.storage.get_binary_obj(
            [self.relic_type, self.name, "html", name]
        ) as f:
            return f.read().decode("utf-8")

    def add_text(self, name: str, text: str) -> None:

        self.assert_valid_id(name)

        size = getsizeof(text)
        metadata = Metadata(
            name=name,
            data_type="text",
            relic_type=self.relic_type,
            storage_name=self.storage.name,
            size=size,
            shape=len(text),
        )

        self.storage.put_text([self.relic_type, self.name, "text", name], text)
        self._add_metadata(metadata)

    def list_text(self) -> List[str]:
        return self.storage.list_keys([self.relic_type, self.name, "text"])

    def get_text(self, name: str) -> str:
        self.assert_valid_id(name)

        return self.storage.get_text([self.relic_type, self.name, "text", name])

    def _add_metadata(self, metadata: Metadata) -> None:
        self.assert_valid_id(metadata.name)

        self.storage.put_metadata(
            [self.relic_type, self.name, "metadata", metadata.data_type, metadata.name],
            metadata.get_dict(),
        )

    def describe(self) -> Dict:
        return self.storage.get_metadata(
            [self.relic_type, self.name, "metadata"], self.name
        )

    """
    Query metadata over all relics

    pass SQL expression as a string
    """

    def query(self, statement: str) -> List:
        metadata = self.storage.get_all_relic_metadata()

        for data in metadata:
            self.metadata_db.sync_metadata(Metadata.parse_dict(data))

        return self.metadata_db.query(statement)

    def add_tag(self, tags: Dict) -> List[Dict]:
        # sync tags

        tag = RelicTag(self.name, self.relic_type, self.storage.name, tags=tags)
        tag.date_created = dt.datetime.utcnow().strftime(dt_format)

        try:
            all_tags = self.metadata_db.get_all_tags_from_relic(
                self.name, self.relic_type, self.storage.name
            )

            saved = self.metadata_db.add_relic_tag(tag)

            all_tags.extend(saved)

            self.storage.put_tags([self.relic_type, self.name, "tags"], all_tags)

            return saved
        except Exception as e:
            logging.warning(f"Error adding tags: {tags} | {e} ")

    def list_tags(self) -> List[str]:
        return self.storage.get_tags([self.relic_type, self.name, "tags"])


class Reliquery:
    """
    Class used to query over available and accessible storage locations and Relics

    Attributes
    ----------
    storages : List[Storage]
        list of availably accessible storages

    metadata_db : MetadataDB
        in memory sqlite db used for querying Relics

    """

    def __init__(self, storages: List[Storage] = []) -> None:
        if len(storages) > 0:
            self.storages = storages
        else:
            self.storages = get_all_available_storages()
        self.metadata_db = MetadataDB()

    def query(self, statement: str) -> List:
        if self.storages:
            self._sync_relics()

        return self.metadata_db.query(statement)


class Reliquery:
    """
    Class used to query over available and accessible storage locations and Relics

    Attributes
    ----------
    storages : List[Storage]
        list of availably accessible storages

    metadata_db : MetadataDB
        in memory sqlite db used for querying Relics

    """

    def __init__(self, storages: List[Storage] = []) -> None:
        if len(storages) > 0:
            self.storages = storages
        else:
            self.storages = get_all_available_storages()
        self.metadata_db = MetadataDB()

    def query(self, statement: str) -> List:
        if self.storages:
            self._sync_relics()

        return self.metadata_db.query(statement)

    def _sync_relics(self) -> None:
        for stor in self.storages:
            metadata = stor.get_all_relic_metadata()
            tags = stor.get_all_relic_tags()

            for data in metadata:
                self.metadata_db.sync_metadata(Metadata.parse_dict(data))

            for tag in tags:

                self.metadata_db.sync_tags(RelicTag.parse_dict(tag))



import sqlite3
from sqlite3 import Error
from sqlite3.dbapi2 import Connection
from typing import Dict, List
import datetime as dt
import logging

dt_format = "%m/%d/%Y %H:%M:%S"


class Data:
    def get_dict(self) -> Dict:
        raise NotImplementedError

    @classmethod
    def parse_dict(self, dict: Dict):
        raise NotImplementedError


class RelicTag(Data):
    def __init__(
        self,
        relic_name: str,
        relic_type: str,
        storage_name: str,
        tags: Dict,
        id: int = None,
        date_created: str = None,
    ) -> None:
        self.id = id
        self.relic_name = relic_name
        self.relic_type = relic_type
        self.storage_name = storage_name
        self.tags = tags
        self.date_created = date_created

    def get_dict(self) -> Dict:
        return {
            "id": self.id,
            "relic_name": self.relic_name,
            "relic_type": self.relic_type,
            "storage_name": self.storage_name,
            "tags": self.tags,
            "date_created": self.date_created,
        }

    @classmethod
    def parse_dict(self, dict: Dict):
        return RelicTag(
            relic_name=dict["relic_name"],
            relic_type=dict["relic_type"],
            storage_name=dict["storage_name"],
            tags=dict["tags"],
            id=dict["id"] if "id" in dict else None,
            date_created=dict["date_created"],
        )

    @classmethod
    def parse_sql_result(self, tag: List):
        return {
            "id": tag[0],
            "relic_name": tag[1],
            "relic_type": tag[2],
            "storage_name": tag[3],
            "tags": {tag[4]: tag[5]},
            "date_created": tag[6],
        }


class Metadata(Data):
    def __init__(
        self,
        name: str,
        data_type: str,
        relic_type: str,
        storage_name: str,
        size: float = None,
        shape: str = None,
        id: int = None,
        last_modified: str = None,
    ) -> None:
        self.id = id
        self.name = name
        self.data_type = data_type
        self.relic_type = relic_type
        self.storage_name = storage_name
        self.size = size
        self.shape = shape
        self.last_modified = (
            last_modified
            if last_modified is not None
            else dt.datetime.utcnow().strftime(dt_format)
        )

    def get_dict(self) -> Dict:
        return {
            "id": self.id,
            "name": self.name,
            "data_type": self.data_type,
            "relic_type": self.relic_type,
            "storage_name": self.storage_name,
            "size": self.size,
            "shape": self.shape,
            "last_modified": self.last_modified,
        }

    @classmethod
    def parse_dict(self, dict: Dict):
        metadata = Metadata(
            name=dict["name"],
            data_type=dict["data_type"],
            relic_type=dict["relic_type"],
            storage_name=dict["storage_name"]
            if "storage_name" in dict
            else "None",  # FIXME
            last_modified=dict["last_modified"],
        )

        if "size" in dict:
            metadata.size = dict["size"]

        if "shape" in dict:
            metadata.shape = dict["shape"]

        return metadata

    @classmethod  # FIXME need correct keys and value indexs
    def parse_sql_results(self, data: List):
        return {
            "id": data[0],
            "relic_name": data[1],
            "relic_type": data[2],
            "storage_name": data[3],
            "tags": data[4],
            "date_created": data[5],
        }


class MetadataDB:

    metadata_table = """
    CREATE TABLE IF NOT EXISTS metadata (
        id integer NOT NULL PRIMARY KEY AUTOINCREMENT,
        name text NOT NULL,
        data_type text NOT NULL,
        relic_type text NOT NULL,
        storage_name text NOT NULL,
        size real,
        shape text,
        last_modified text NOT NULL
    )
    """

    relic_tag_table = """
    CREATE TABLE IF NOT EXISTS relic_tags (
        id integer NOT NULL PRIMARY KEY AUTOINCREMENT,
        relic_name text NOT NULL,
        relic_type text NOT NULL,
        storage_name text NOT NULL,
        key text NOT NULL,
        value text NOT NULL,
        date_created text NOT NULL
    )
    """

    def __init__(self) -> None:
        self.conn = None

        try:
            self.conn = self.connect_db()
        except Error as e:
            logging.warning(f"Error connecting to database: {e}")
        finally:
            if self.conn:
                try:
                    cur = self.conn.cursor()
                    cur.execute(self.metadata_table)
                    cur.execute(self.relic_tag_table)
                    self.conn.commit()
                except Error as e:
                    logging.warning(f"Error creating database tables: {e}")

    def connect_db(self) -> Connection:
        return sqlite3.connect(":memory:")

    def get_db_filename(self) -> str:
        return self.file_loc

    def add_metadata(self, metadata: Metadata) -> None:
        try:
            cur = self.conn.cursor()

            cur.execute(
                """
                SELECT * FROM metadata 
                WHERE name = ?
                AND
                data_type = ?
                AND
                relic_type = ?
                AND
                storage_name = ?
                LIMIT 1
                """,
                (
                    metadata.name,
                    metadata.data_type,
                    metadata.relic_type,
                    metadata.storage_name,
                ),
            )
            rows = cur.fetchall()

            if len(rows) > 0:
                self.update_metadata(metadata)
                return

            cur.execute(
                "INSERT into metadata VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    None,
                    metadata.name,
                    metadata.data_type,
                    metadata.relic_type,
                    metadata.storage_name,
                    metadata.size,
                    metadata.shape,
                    metadata.last_modified,
                ),
            )

            self.conn.commit()

        except Error as e:
            print(e)

    def get_metadata_by_name(
        self, name: str, data_type: str, relic_type: str, storage_name: str
    ) -> Metadata:
        try:
            cur = self.conn.cursor()

            cur.execute(
                """
                SELECT * FROM metadata 
                WHERE name = ?
                AND
                data_type = ?
                AND
                relic_type = ?
                AND
                storage_name = ?
                LIMIT 1
                """,
                (name, data_type, relic_type, storage_name),
            )
            rows = cur.fetchall()

            if len(rows) > 0:
                return Metadata(
                    name=rows[0][1],
                    data_type=rows[0][2],
                    relic_type=rows[0][3],
                    storage_name=rows[0][4],
                    size=rows[0][5],
                    last_modified=rows[0][7],
                    shape=rows[0][6],
                    id=rows[0][0],
                )
            else:
                return None

        except Error as e:
            print(e)

    def get_all_metadata(self) -> List:
        try:
            cur = self.conn.cursor()

            cur.execute("""SELECT * FROM metadata """)
            rows = cur.fetchall()
            print(rows)
            if len(rows) > 0:
                for row in rows:
                    yield Metadata(
                        name=row[1],
                        data_type=row[2],
                        relic_type=row[3],
                        storage_name=row[4],
                        size=row[5],
                        last_modified=row[7],
                        shape=row[6],
                        id=row[0],
                    )
            else:
                return []

        except Error as e:
            print(e)

    def update_metadata(self, metadata: Metadata) -> None:
        try:
            cur = self.conn.cursor()

            cur.execute(
                """
                UPDATE metadata 
                SET name=?, data_type=?, relic_type=?, storage_name=?, size=?, shape=?, last_modified=?
                WHERE id=?
                """,
                (
                    metadata.name,
                    metadata.data_type,
                    metadata.relic_type,
                    metadata.storage_name,
                    metadata.size,
                    metadata.shape,
                    metadata.last_modified,
                    metadata.id,
                ),
            )

            self.conn.commit()

        except Error as e:
            print(e)

    def delete_metadata(self, name: str) -> None:
        raise RuntimeError("method not initialized")

    def query(self, statement: str) -> List:
        results = []

        try:
            cur = self.conn.cursor()

            cur.execute(statement)

            results = cur.fetchall()
        except Error as e:
            logging.warning(f"Error quering database: {e}")

        return results

    def sync_metadata(self, ext: Metadata) -> None:
        int = self.get_metadata_by_name(
            ext.name, ext.data_type, ext.relic_type, ext.storage_name
        )

        if int is not None:
            if dt.datetime.strptime(
                ext.last_modified, dt_format
            ) > dt.datetime.strptime(int.last_modified, dt_format):
                self.update_metadata(ext)
        else:
            self.add_metadata(ext)

    def sync_tags(self, ext: RelicTag) -> None:
        if len(self.get_by_relic_tag(ext)) == 0:
            self.add_relic_tag(
                ext
            )  # FIXME something here is breaking querying reliquery

    def add_relic_tag(self, relic_tag: RelicTag) -> List[RelicTag]:
        try:
            cur = self.conn.cursor()

            for key, value in relic_tag.tags.items():
                cur.execute(
                    """
                    INSERT INTO relic_tags VALUES (?,?,?,?,?,?,?)
                    """,
                    (
                        relic_tag.id,
                        relic_tag.relic_name,
                        relic_tag.relic_type,
                        relic_tag.storage_name,
                        key,
                        value,
                        relic_tag.date_created,
                    ),
                )

            self.conn.commit()
        except Error as e:
            logging.warning(f"Error creating relic tag: {relic_tag.get_dict()} | {e}")

        return self.get_by_relic_tag(relic_tag)

    def remove_relic_tag(self, relic_tag: RelicTag) -> None:
        try:
            cur = self.conn.cursor()

            for key, value in relic_tag.tags:
                cur.execute(
                    """
                    DELETE FROM relic_tags
                    WHERE relic_name=?
                    AND relic_type=?
                    AND storage_type=?
                    AND key=?
                    AND value=?
                    """,
                    (
                        relic_tag.relic_name,
                        relic_tag.relic_type,
                        relic_tag.storage_name,
                        key,
                        value,
                    ),
                )

            self.conn.commit()

        except Error as e:
            logging.warning(f"Error deleting relic tag: {e}")

    def get_by_key_value(self, key: str, value: str) -> List:
        tags = []
        try:
            cur = self.conn.cursor()
            cur.execute(
                """
                SELECT * FROM relic_tags
                WHERE key=?
                AND value=?
                """,
                (key, value),
            )
            tags.extend(cur.fetchall())

        except Error as e:
            logging.warning(f"Error getting relic tag: {e}")

        return tags

    def get_all_tags_from_relic(
        self, relic_name: str, relic_type: str, storage_name: str
    ) -> List:
        tags = []

        try:
            cur = self.conn.cursor()

            cur.execute(
                """
                SELECT * FROM relic_tags
                WHERE relic_name=?
                AND relic_type=?
                AND storage_name=?
                """,
                (relic_name, relic_type, storage_name),
            )

            queries = cur.fetchall()

            if queries:
                for tag in queries:
                    tags.append(RelicTag.parse_sql_result(tag))

        except Error as e:
            logging.warning(f"Error getting all tags: {e}")

        return tags

    def get_by_relic_tag(self, tag: RelicTag) -> List:
        tags = []
        try:
            cur = self.conn.cursor()

            for key, value in tag.tags.items():
                cur.execute(
                    """
                    SELECT * FROM relic_tags
                    WHERE key=?
                    AND value=?
                    AND relic_name=?
                    AND relic_type=?
                    AND storage_name=?
                    """,
                    (key, value, tag.relic_name, tag.relic_type, tag.storage_name),
                )
                tags.extend(cur.fetchall())

                self.conn.commit()

        except Error as e:
            logging.warning(f"Error getting relic tag: {e}")

        return list(map(RelicTag.parse_sql_result, tags))



import os
import json


def get_config(reliquery_dir):
    # First we check the environment for the config
    if "RELIQUERY_CONFIG" in os.environ:
        return json.loads(os.environ["RELIQUERY_CONFIG"])

    # Otherwise, fallback on a file.
    config_path = os.path.join(reliquery_dir, "config")

    if os.path.exists(config_path):
        with open(config_path, mode="r") as config_file:
            config = json.load(config_file)
        return config

    return {
        "default": {
            "storage": {
                "type": "File",
                "args": {
                    "root": reliquery_dir,
                },
            }
        },
        "demo": {
            "storage": {
                "type": "S3",
                "args": {
                    "s3_signed": False,
                    "s3_bucket": "reliquery",
                    "prefix": "relics",
                },
            }
        },
    }



import logging
import os
from io import BytesIO, BufferedIOBase
from posixpath import join
from typing import Any, List, Dict
from shutil import copyfile
import json

import boto3
from botocore import UNSIGNED
from botocore.client import Config
from numpy.lib.function_base import append


from . import settings
from reliquery.metadata import Metadata, MetadataDB

import reliquery

StoragePath = List[str]


class StorageItemDoesNotExist(Exception):
    pass


class Storage:
    def put_file(self, path: StoragePath, file_path: str) -> None:
        raise NotImplementedError

    def put_binary_obj(self, path: StoragePath, buffer: BytesIO):
        raise NotImplementedError

    def put_binary_obj(self, path: StoragePath, buffer: BytesIO):
        raise NotImplementedError

    def get_binary_obj(self, path: StoragePath) -> BytesIO:
        raise NotImplementedError

    def put_text(self, path: StoragePath, buffer: BytesIO) -> None:
        raise NotImplementedError

    def get_text(self, path: StoragePath) -> BytesIO:
        raise NotImplementedError

    def list_keys(self, path: StoragePath) -> List[str]:
        raise NotImplementedError

    def put_metadata(self, path: StoragePath, metadata: Dict):
        raise NotImplementedError

    def get_metadata(self, path: StoragePath, root_key: str) -> Dict:
        raise NotImplementedError

    def get_all_relic_metadata(self) -> List[Dict]:
        raise NotImplementedError

    def put_tags(self, path: StoragePath, tags: List[Dict]) -> None:
        raise NotImplementedError

    def get_tags(self, path: StoragePath) -> List[Dict]:
        raise NotImplementedError

    def get_all_relic_tags(self) -> List[Dict]:
        raise NotImplementedError


class FileStorage:
    def __init__(self, root: str, name: str):
        self.root = root
        self.name = name

    def _ensure_path(self, path: StoragePath):
        if not os.path.exists(self._join_path(path[:-1])):
            os.makedirs(self._join_path(path[:-1]), exist_ok=True)

    def _join_path(self, path: StoragePath) -> str:
        return os.path.join(*[self.root] + path)

    def put_file(self, path: StoragePath, file_path: str) -> None:
        self._ensure_path(path)
        copyfile(file_path, self._join_path(path))

    def put_binary_obj(self, path: StoragePath, buffer: BytesIO) -> None:
        self._ensure_path(path)
        with open(self._join_path(path), "wb") as f:
            f.write(buffer.getbuffer())

    def get_binary_obj(self, path: StoragePath) -> BytesIO:
        return open(self._join_path(path), "rb")

    def put_text(self, path: StoragePath, text: str) -> None:
        self._ensure_path(path)

        with open(self._join_path(path), "w") as f:
            return f.write(text)

    def get_text(self, path: StoragePath) -> str:
        try:
            with open(self._join_path(path), "r") as f:
                return f.read()
        except IOError:
            raise StorageItemDoesNotExist

    def list_keys(self, path: StoragePath) -> List[str]:
        joined_path = self._join_path(path)
        if not os.path.exists(joined_path):
            return []

        if not os.path.isdir(joined_path):
            return []

        return os.listdir(joined_path)

    def put_metadata(self, path: StoragePath, metadata: Dict):
        self._ensure_path(path)

        with open(self._join_path(path), "w") as f:
            return f.write(json.dumps(metadata))

    def get_metadata(self, path: StoragePath, root_key: str) -> Dict:
        data = {root_key: {"arrays": [], "text": [], "html": []}}

        dirs = ["arrays", "html", "text"]

        def dict_from_path(path: StoragePath, dirname: str):
            dirpath = path.copy()
            dirpath.append(dirname)
            if os.path.exists(self._join_path(dirpath)):
                entries = os.listdir(self._join_path(dirpath))
                for i in entries:
                    entry_path = dirpath.copy()
                    entry_path.append(i)
                    with open(self._join_path(entry_path), "r") as f:
                        data[root_key][dirname].append(json.loads(f.read()))

        for d in dirs:
            dict_from_path(path, d)

        return data

    # TODO Test coverage needed
    def get_all_relic_metadata(self, encoding: str = "utf-8") -> List[Dict]:
        meta_keys = [
            key for key in self.list_key_paths([""]) if "metadata" in key.split("/")
        ]

        if len(meta_keys) > 0:
            for key in meta_keys:
                with open(key, "r") as f:
                    yield json.loads(f.read())

    def list_key_paths(self, path: StoragePath) -> List[str]:
        paths = []
        joined_path = self._join_path(path)
        subs = os.listdir(joined_path)

        if subs:
            for sub in subs:
                copy = path.copy()
                copy.append(sub)
                sub_path = self._join_path(copy)
                if os.path.isdir(sub_path):
                    paths.extend(self.list_key_paths(copy))
                else:
                    paths.append(sub_path)

        return paths

    def put_tags(self, path: StoragePath, tags: List[Dict]) -> None:
        self._ensure_path(path)
        jsons = [json.dumps(tag) for tag in tags]

        with open(self._join_path(path), "w") as f:
            return f.write(";".join(jsons))

    def get_tags(self, path: StoragePath) -> List[Dict]:
        self._ensure_path(path)

        with open(self._join_path(path), "r") as f:
            return list(map(json.loads, f.read().split(";")))

    def get_all_relic_tags(self) -> List[Dict]:
        tag_keys = [
            key for key in self.list_key_paths([""]) if "tags" in key.split("/")
        ]

        for key in tag_keys:
            return self.get_tags(key.split("/")[-3:])


S3Client = Any


def _get_s3_client(signed) -> S3Client:
    if signed:
        return boto3.client("s3")
    else:
        return boto3.client("s3", config=Config(signature_version=UNSIGNED))


class S3Storage(Storage):
    def __init__(
        self,
        s3_bucket: str,
        prefix: str,
        name: str,
        s3_signed: bool = True,
    ):
        self.s3_bucket = s3_bucket
        self.prefix = prefix
        self.name = name if len(name) > 0 else s3_bucket
        self.signed = s3_signed

        self.s3 = _get_s3_client(self.signed)

    def _join_path(self, path: StoragePath) -> str:
        return "/".join([self.prefix] + path)

    def put_file(self, path: StoragePath, file_path: str) -> None:
        self.s3.upload_file(file_path, self.s3_bucket, self._join_path(path))

    def put_binary_obj(self, path: StoragePath, buffer: BufferedIOBase) -> None:
        print(buffer)

        self.s3.upload_fileobj(buffer, self.s3_bucket, self._join_path(path))

    def get_binary_obj(self, path: StoragePath) -> BufferedIOBase:
        buffer = BytesIO()

        self.s3.download_fileobj(self.s3_bucket, self._join_path(path), buffer)

        buffer.seek(0)

        return buffer

    def put_text(self, path: StoragePath, text: str, encoding: str = "utf-8") -> None:
        self.s3.put_object(
            Key=self._join_path(path), Bucket=self.s3_bucket, Body=text.encode(encoding)
        )

    def get_text(self, path: StoragePath, encoding: str = "utf-8") -> str:
        try:
            obj = self.s3.get_object(Key=self._join_path(path), Bucket=self.s3_bucket)
        except self.s3.exceptions.NoSuchKey:
            raise StorageItemDoesNotExist

        return obj["Body"].read().decode(encoding)

    def list_keys(self, path: StoragePath) -> List[str]:
        prefix = self._join_path(path)

        def process_key(k):
            return k[len(prefix) + 1 :]

        is_truncated = True
        keys = []
        kwargs = {}
        while is_truncated:
            response = self.s3.list_objects_v2(
                Bucket=self.s3_bucket, Prefix=prefix, **kwargs
            )

            is_truncated = response["IsTruncated"]
            if is_truncated:
                kwargs = dict(ContinuationToken=response["NextContinuationToken"])

            try:
                keys.extend([process_key(c["Key"]) for c in response["Contents"]])
            except KeyError as e:
                logging.warning(f"No files found in directory {prefix}")
        # we want to remove the prefx
        return keys

    def put_metadata(self, path: Storage, metadata: Dict):
        self.s3.put_object(
            Key=self._join_path(path), Bucket=self.s3_bucket, Body=json.dumps(metadata)
        )

    def get_metadata(self, path: StoragePath, root_key: str) -> Dict:
        dirs = ["arrays", "html", "text"]

        data = {root_key: {"arrays": [], "text": [], "html": []}}

        def dict_from_path(path: StoragePath, dirname: str) -> None:
            dirpath = path.copy()
            dirpath.append(dirname)
            dir_keys = self.list_keys(dirpath.copy())

            if len(dir_keys) > 0:
                for key in dir_keys:
                    key_path = dirpath.copy()
                    key_path.append(key)

                    try:
                        obj = self.s3.get_object(
                            Key=self._join_path(key_path),
                            Bucket=self.s3_bucket,
                        )
                    except self.s3.exceptions.NoSuchKey:
                        raise StorageItemDoesNotExist

                    data[root_key][dirname].append(
                        json.loads(obj["Body"].read().decode("utf-8"))
                    )

        for d in dirs:
            dict_from_path(path, d)

        return data

    def get_all_relic_metadata(self, encoding: str = "utf-8") -> List[Dict]:
        meta_keys = [
            key for key in self.list_key_paths([""]) if "metadata" in key.split("/")
        ]

        if len(meta_keys) > 0:
            for key in meta_keys:
                yield json.loads(
                    self.s3.get_object(
                        Key=self._join_path(key.split("/")),
                        Bucket=self.s3_bucket,
                    )["Body"]
                    .read()
                    .decode(encoding)
                )

    def list_key_paths(self, path: StoragePath) -> List[str]:
        prefix = self._join_path(path)

        def process_key(k):
            return k[len(prefix) :]

        is_truncated = True
        keys = []
        kwargs = {}
        while is_truncated:
            response = self.s3.list_objects_v2(
                Bucket=self.s3_bucket, Prefix=prefix, **kwargs
            )

            is_truncated = response["IsTruncated"]
            if is_truncated:
                kwargs = dict(ContinuationToken=response["NextContinuationToken"])

            try:
                keys.extend([process_key(c["Key"]) for c in response["Contents"]])
            except KeyError as e:
                logging.warning(f"No files found in directory {prefix}")
        # we want to remove the prefx
        return keys

    def put_tags(self, path: StoragePath, tags: List[Dict]) -> None:
        jsons = [json.dumps(tag) for tag in tags]

        self.put_text(path, ";".join(jsons))

    def get_tags(self, path: StoragePath) -> List[Dict]:
        return list(map(json.loads, self.get_text(path).split(";")))

    def get_all_relic_tags(self) -> List[Dict]:
        tag_keys = [
            key for key in self.list_key_paths([""]) if "tags" in key.split("/")
        ]
        tags = []
        for key in tag_keys:
            tags.extend(self.get_tags(key.split("/")))

        return tags


def get_storage_by_name(name: str, root: str = os.path.expanduser("~")) -> Storage:
    reliquery_dir = os.path.join(root, "reliquery")
    config = settings.get_config(reliquery_dir)

    return get_storage(name, reliquery_dir, config[name])


def get_all_available_storages(root: str = os.path.expanduser("~")) -> List[Storage]:
    reliquery_dir = os.path.join(root, "reliquery")
    config = settings.get_config(reliquery_dir)

    storages = []

    for key in config.keys():
        storages.append(get_storage(key, reliquery_dir, config[key]))

    return storages


def get_storage(name: str, root: str, config: Dict) -> Storage:
    if config["storage"]["type"] == "S3":
        return S3Storage(
            **config["storage"]["args"],
            name=name,
        )

    elif config["storage"]["type"] == "File":
        if "root" in config["storage"]["args"]:
            return FileStorage(**config["storage"]["args"], name=name)
        else:
            return FileStorage(**config["storage"]["args"], root=root, name=name)

    else:
        raise ValueError(f"No storage found by name : {name}")



from typing import Set

from io import BytesIO
from .property import Serializer

try:
    import numpy as np

    has_numpy = True
except:
    has_numpy = False

if has_numpy:

    class Arrays(Serializer[np.ndarray]):
        programmatic_name = "arrays"

        def serialize(self, item: np.ndarray):
            buffer = BytesIO()
            raw_data = np.save(buffer, item, allow_pickle=False)

            buffer.seek(0)

            return buffer

        def deserialize(self, buffer: BytesIO) -> np.ndarray:
            return np.load(buffer, buffer, allow_pickle=False)

        def __setitem__(self, key: str, item: np.ndarray):
            # Serialize the data.
            buffer = BytesIO()
            raw_data = np.save(buffer, item, allow_pickle=False)

            buffer.seek(0)

            # Push the data.
            self.relic.upload_data("arrays", key)

        def __getitem__(self, key: str) -> np.ndarray:
            pass

        def keys(self) -> Set[str]:
            return set()
